from __future__ import annotations

from types import SimpleNamespace
from typing import Optional, Union, Callable
from joblib import Parallel, delayed

import numpy as np
from pydantic import NegativeInt, NonNegativeInt, PositiveInt, confloat
from numba import njit, prange
from sklearn.utils import check_random_state

from sequentia.models.knn.base import _KNNMixin, _KNNValidator
from sequentia.models.base import _Classifier

from sequentia.utils.decorators import _validate_params, _requires_fit, _override_params
from sequentia.utils.data import SequentialDataset
from sequentia.utils.multiprocessing import _effective_n_jobs
from sequentia.utils.validation import (
    _check_classes,
    Array,
    _MultivariateFloatSequenceClassifierValidator
)

__all__ = ['KNNClassifier']

_defaults = SimpleNamespace(
    **{
        **_KNNMixin._defaults.__dict__,
        "classes": None,
    }
)


class _KNNClassifierValidator(_KNNValidator):
    classes: Optional[Array[int]] = _defaults.classes


class KNNClassifier(_KNNMixin, _Classifier):
    """A k-nearest neighbor classifier that uses DTW as a distance measure for sequence comparison.

    The classifier computes the score for each class as the total of the distance weightings of every sequence belonging to that class,
    within the DTW k-neighborhood of the sequence being classified.

    Examples
    --------
    Using a :class:`.KNNClassifier` to classify spoken digits. ::

        import numpy as np
        from sequentia.datasets import load_digits
        from sequentia.models.knn import KNNClassifier

        # Seed for reproducible pseudo-randomness
        random_state = np.random.RandomState(1)

        # Fetch MFCCs of spoken digits
        data = load_digits()
        train_data, test_data = data.split(test_size=0.2, random_state=random_state)

        # Create a HMMClassifier using a class frequency prior
        clf = KNNClassifier()

        # Fit the classifier
        X_train, y_train, lengths_train = train_data.X_y_lengths
        clf.fit(X_train, y_train, lengths_train)

        # Predict classes for the test observation sequences
        X_test, lengths_test = test_data.X_lengths
        y_pred = clf.predict(X_test, lengths_test)
    """

    _defaults = _defaults


    @_validate_params(using=_KNNClassifierValidator)
    def __init__(
        self,
        *,
        k: PositiveInt = _defaults.k,
        weighting: Optional[Callable] = _defaults.weighting,
        window: confloat(ge=0, le=1) = _defaults.window,
        independent: bool = _defaults.independent,
        classes: Optional[Array[int]] = None,
        use_c: bool = _defaults.use_c,
        n_jobs: Union[NegativeInt, PositiveInt] = _defaults.n_jobs,
        random_state: Optional[Union[NonNegativeInt, np.random.RandomState]] = _defaults.random_state
    ) -> KNNClassifier:
        """Initializes the :class:`.KNNClassifier`.

        :param k: Number of neighbors.

        :param weighting: A callable that specifies how distance weighting should be performed.
            The callable should accept a :class:`numpy:numpy.ndarray` of DTW distances, apply an element-wise weighting transformation
            to the matrix of DTW distances, then return an equally-sized :class:`numpy:numpy.ndarray` of weightings.
            If ``None``, then a uniform weighting of 1 will be applied to all distances.

        :param window: The size of the Sakoe—Chiba band global constrant as a fraction of the length of the shortest of the two sequences being compared.

            - A larger window will give more freedom to the DTW alignment, allowing more deviation but leading to potentially slower computation.
              A window of 1 is equivalent to full DTW computation with no global constraint applied.
            - A smaller window will restrict the DTW alignment, and possibly speed up the DTW computation.
              A window of 0 is equivalent to Euclidean distance.

        :param independent: Whether or not to allow features to be warped independently from each other. See [#dtw_multi]_ for an overview of independent and dependent dynamic time warping.

        :param classes: Set of possible class labels.

            - If not provided, these will be determined from the training data labels.
            - If provided, output from methods such as :func:`predict_proba` and :func:`predict_scores`
              will follow the ordering of the class labels provided here.

        :param use_c: Whether or not to use fast pure C compiled functions from `dtaidistance <https://github.com/wannesm/dtaidistance>`__ to perform the DTW computations.

        :param n_jobs: Maximum number of concurrently running workers.

            - If 1, no parallelism is used at all (useful for debugging).
            - If -1, all CPUs are used.
            - If < -1, ``(n_cpus + 1 + n_jobs)`` are used — e.g. ``n_jobs=-2`` uses all but one.

        :param random_state: Seed or :class:`numpy:numpy.random.RandomState` object for reproducible pseudo-randomness.
        """
        #: Number of neighbors.
        self.k = k
        #: A callable that specifies how distance weighting should be performed.
        self.weighting = weighting
        #: The size of the Sakoe—Chiba band global constrant as a fraction of the length of the shortest of the two sequences being compared.
        self.window = window
        #: Whether or not to allow features to be warped independently from each other.
        self.independent = independent
        #: Set of possible class labels.
        self.classes = classes
        #: Whether or not to use fast pure C compiled functions from `dtaidistance <https://github.com/wannesm/dtaidistance>`__ to perform the DTW computations.
        self.use_c = use_c
        #: Maximum number of concurrently running workers.
        self.n_jobs = n_jobs
        #: Seed or :class:`numpy:numpy.random.RandomState` object for reproducible pseudo-randomness.
        self.random_state = random_state


    def fit(
        self,
        X: Array[float],
        y: Array[int],
        lengths: Optional[Array[int]] = None
    ) -> KNNClassifier:
        """Fits the classifier to the sequence(s) in ``X``.

        :param X: Univariate or multivariate observation sequence(s).

            - Should be a single 1D or 2D array.
            - Should have length as the 1st dimension and features as the 2nd dimension.
            - Should be a concatenated sequence if multiple sequences are provided,
              with respective sequence lengths being provided in the ``lengths`` argument for decoding the original sequences.

        :param y: Classes corresponding to sequence(s) provided in ``X``.

        :param lengths: Lengths of the observation sequence(s) provided in ``X``.

            - If ``None``, then ``X`` is assumed to be a single observation sequence.
            - ``len(X)`` should be equal to ``sum(lengths)``.

        :return: The fitted classifier.
        """
        data = _MultivariateFloatSequenceClassifierValidator(X=X, y=y, lengths=lengths)
        self.X_ = data.X
        self.y_ = data.y
        self.lengths_ = data.lengths
        self.idxs_ = SequentialDataset._get_idxs(data.lengths)
        self.random_state_ = check_random_state(self.random_state)
        self.classes_ = _check_classes(data.y, self.classes)
        return self


    @_requires_fit
    def predict(
        self,
        X: Array[float],
        lengths: Optional[Array[int]] = None
    ) -> Array[int]:
        """Predicts classes for the sequence(s) in ``X``.

        :param X: Univariate or multivariate observation sequence(s).

            - Should be a single 1D or 2D array.
            - Should have length as the 1st dimension and features as the 2nd dimension.
            - Should be a concatenated sequence if multiple sequences are provided,
              with respective sequence lengths being provided in the ``lengths`` argument for decoding the original sequences.

        :param lengths: Lengths of the observation sequence(s) provided in ``X``.

            - If ``None``, then ``X`` is assumed to be a single observation sequence.
            - ``len(X)`` should be equal to ``sum(lengths)``.

        :note: This method requires a trained classifier — see :func:`fit`.

        :return: Class predictions.
        """
        class_scores = self.predict_scores(X, lengths)
        return self._find_max_labels(class_scores)


    def fit_predict(
        self,
        X: Array[float],
        y: Array[int],
        lengths: Optional[Array[int]] = None
    ) -> Array[int]:
        """Fits the classifier to the sequence(s) in ``X`` and predicts classes for ``X``.

        :param X: Univariate or multivariate observation sequence(s).

            - Should be a single 1D or 2D array.
            - Should have length as the 1st dimension and features as the 2nd dimension.
            - Should be a concatenated sequence if multiple sequences are provided,
              with respective sequence lengths being provided in the ``lengths`` argument for decoding the original sequences.

        :param y: Classes corresponding to sequence(s) provided in ``X``.

        :param lengths: Lengths of the observation sequence(s) provided in ``X``.

            - If ``None``, then ``X`` is assumed to be a single observation sequence.
            - ``len(X)`` should be equal to ``sum(lengths)``.

        :return: Class predictions.
        """
        return super().fit_predict(X, y, lengths)


    @_requires_fit
    def predict_proba(
        self,
        X: Array[float],
        lengths: Optional[Array[int]] = None
    ) -> Array[float]:
        """Predicts class probabilities for the sequence(s) in ``X``.

        Probabilities are calculated as normalized class scores.

        :param X: Univariate or multivariate observation sequence(s).

            - Should be a single 1D or 2D array.
            - Should have length as the 1st dimension and features as the 2nd dimension.
            - Should be a concatenated sequence if multiple sequences are provided,
              with respective sequence lengths being provided in the ``lengths`` argument for decoding the original sequences.

        :param lengths: Lengths of the observation sequence(s) provided in ``X``.

            - If ``None``, then ``X`` is assumed to be a single observation sequence.
            - ``len(X)`` should be equal to ``sum(lengths)``.

        :note: This method requires a trained classifier — see :func:`fit`.

        :return: Class membership probabilities.
        """
        class_scores = self.predict_scores(X, lengths)
        return class_scores / class_scores.sum(axis=1, keepdims=True)


    @_requires_fit
    def predict_scores(
        self,
        X: Array[float],
        lengths: Optional[Array[int]] = None
    ) -> Array[float]:
        """Predicts class scores for the sequence(s) in ``X``.

        Scores are calculated as the class distance weighting sums of all training sequences in the k-neighborhood.

        :param X: Univariate or multivariate observation sequence(s).

            - Should be a single 1D or 2D array.
            - Should have length as the 1st dimension and features as the 2nd dimension.
            - Should be a concatenated sequence if multiple sequences are provided,
              with respective sequence lengths being provided in the ``lengths`` argument for decoding the original sequences.

        :param lengths: Lengths of the observation sequence(s) provided in ``X``.

            - If ``None``, then ``X`` is assumed to be a single observation sequence.
            - ``len(X)`` should be equal to ``sum(lengths)``.

        :note: This method requires a trained classifier — see :func:`fit`.

        :return: Class scores.
        """
        _, k_distances, k_labels = self.query_neighbors(X, lengths, sort=False)
        k_weightings = self._weighting()(k_distances)
        return self._compute_scores(k_labels, k_weightings)


    @_requires_fit
    def score(
        self,
        X: Array,
        y: Array[int],
        lengths: Optional[Array[int]],
        normalize: bool = True,
        sample_weight: Optional[Array] = None,
    ) -> float:
        """Calculates accuracy for the sequence(s) in ``X``.

        :param X: Univariate or multivariate observation sequence(s).

            - Should be a single 1D or 2D array.
            - Should have length as the 1st dimension and features as the 2nd dimension.
            - Should be a concatenated sequence if multiple sequences are provided,
              with respective sequence lengths being provided in the ``lengths`` argument for decoding the original sequences.

        :param y: Classes corresponding to the observation sequence(s) in ``X``.

        :param lengths: Lengths of the observation sequence(s) provided in ``X``.

            - If ``None``, then ``X`` is assumed to be a single observation sequence.
            - ``len(X)`` should be equal to ``sum(lengths)``.

        :param normalize: See :func:`sklearn:sklearn.metrics.accuracy_score`.

        :param sample_weight: See :func:`sklearn:sklearn.metrics.accuracy_score`.

        :note: This method requires a trained classifier — see :func:`fit`.

        :return: Classification accuracy.
        """
        return super().score(X, y, lengths, normalize, sample_weight)


    @_validate_params(using=_KNNValidator)
    @_override_params(_KNNValidator.fields(), temporary=False)
    def set_params(self, **kwargs):
        return self


    def _compute_scores(
        self,
        labels: Array[int],
        weightings: Array[float]
    ) -> Array[float]:
        """Calculates the sum of the weightings for each label group.

        TODO
        """
        scores = np.zeros((len(labels), len(self.classes_)))
        for i, k in enumerate(self.classes_):
            scores[:, i] = np.einsum('ij,ij->i', labels == k, weightings)
        return scores


    def _find_max_labels(
        self,
        scores: Array[float]
    ) -> Array[int]:
        """Returns the label of the k nearest neighbors with the highest score for each example.

        TODO
        """
        n_jobs = _effective_n_jobs(self.n_jobs, scores)
        score_chunks = np.array_split(scores, n_jobs)
        return np.concatenate(
            Parallel(n_jobs=n_jobs, max_nbytes=None)(
                delayed(self._find_max_labels_chunk)(score_chunk)
                for score_chunk in score_chunks
            )
        )


    def _find_max_labels_chunk(
        self,
        score_chunk: Array[float]
    ) -> Array[int]:
        """Returns the label with the highest score for each item in the chunk.

        TODO
        """
        max_labels = np.zeros(len(score_chunk), dtype=int)
        for i, scores in enumerate(score_chunk):
            max_score_idxs = self._multi_argmax(scores)
            max_labels[i] = self.random_state_.choice(self.classes_[max_score_idxs], size=1)
        return max_labels


    @staticmethod
    @njit
    def _multi_argmax(
        arr: Array
    ) -> Array[int]:
        """Same as numpy.argmax but returns all occurrences of the maximum and only requires a single pass.
        From: https://stackoverflow.com/a/58652335

        TODO
        """
        all_, max_ = [0], arr[0]
        for i in prange(1, len(arr)):
            if arr[i] > max_:
                all_, max_ = [i], arr[i]
            elif arr[i] == max_:
                all_.append(i)
        return np.array(all_)



import os
import math
import pytest
from copy import deepcopy
from tempfile import TemporaryDirectory
from unittest.mock import Mock

import numpy as np

from sequentia.datasets import load_digits
from sequentia.models.knn import KNNRegressor
from sequentia.utils.validation import _check_is_fitted

from ....support.assertions import assert_equal

n_classes = 3


@pytest.fixture(scope='module')
def random_state(request):
    return np.random.RandomState(1)


@pytest.fixture(scope='module')
def dataset():
    return load_digits(digits=range(n_classes))


def assert_fit(clf, data):
    assert hasattr(clf, 'X_')
    assert hasattr(clf, 'y_')
    assert hasattr(clf, 'lengths_')
    assert hasattr(clf, 'idxs_')
    assert _check_is_fitted(clf, return_=True)
    assert_equal(clf.X_, data.X)
    assert_equal(clf.y_, data.y)
    assert_equal(clf.lengths_, data.lengths)


@pytest.mark.parametrize('k', [1, 2, 5])
@pytest.mark.parametrize('weighting', [None, lambda x: np.exp(-x)])
def test_regressor_e2e(request, k, weighting, dataset, random_state):
    clf = KNNRegressor(k=k, weighting=weighting, random_state=random_state)

    assert clf.k == k
    assert clf.weighting == weighting

    data = dataset.copy()
    data._X = data._X[:, :1] # only use one feature
    subset, _ = data.split(test_size=0.98, random_state=random_state, stratify=True)
    train, test = subset.split(test_size=0.2, random_state=random_state, stratify=True)

    assert_fit(clf.fit(*train.X_y_lengths), train)
    params = clf.get_params()

    y_pred = clf.predict(*test.X_lengths)
    assert np.issubdtype(y_pred.dtype, np.floating)
    assert y_pred.shape == (len(test),)

    acc = clf.score(*test.X_y_lengths)

    # check serialization/deserialization
    with TemporaryDirectory() as temp_dir:
        model_path = f"{temp_dir}/{request.node.originalname}.model"
        # check that save works
        clf.save(model_path)
        assert os.path.isfile(model_path)
        # check that load works
        clf = KNNRegressor.load(model_path)
        assert (set(clf.get_params().keys()) - set(['weighting'])) == (set(params.keys()) - set(['weighting']))
        # sanity check that custom weighting functions are the same
        if weighting:
            x = random_state.rand(100)
            assert_equal(weighting(x), clf.weighting(x))
        # check that loaded model is fitted and can make predictions
        assert_fit(clf, train)
        y_pred_load = clf.predict(*test.X_lengths)
        if k == 1:
            # predictions should be same as before
            assert_equal(y_pred, y_pred_load)


def test_regressor_predict_train(dataset, random_state):
    """Should be able to perfectly predict training data with k=1"""
    clf = KNNRegressor(k=1, random_state=random_state)

    data = dataset.copy()
    data._X = data._X[:, :1] # only use one feature
    train, _ = data.split(train_size=0.05, random_state=random_state, stratify=True)

    assert_fit(clf.fit(*train.X_y_lengths), train)
    assert math.isclose(clf.score(*train.X_y_lengths), 1.)


def test_regressor_weighting(random_state):
    clf = KNNRegressor(k=3, weighting=lambda x: np.where(x > 10, 0.5, 1))
    clf.random_state_ = random_state

    clf.query_neighbors = Mock(
        return_value=(
            None,
            np.array([
                [1.5, 2, 1],
                [2.5, 1, 0.5],
            ]),
            np.array([
                [10.2, 11.5, 10.4],
                [8.0, 6.5, 5.5]
            ])
        )
    )

    assert_equal(
        clf.predict(None, None),
        np.array([
            (10.2 * 0.5 + 11.5 * 0.5 + 10.4 * 0.5) / (0.5 * clf.k),
            (8.0 * 1 + 6.5 * 1 + 5.5 * 1) / (1 * clf.k)
        ])
    )


@pytest.mark.parametrize('k', [1, 2, 5])
@pytest.mark.parametrize('sort', [True, False])
def test_regressor_query_neighbors(k, sort, dataset, random_state):
    clf = KNNRegressor(k=k, random_state=random_state)

    data = dataset.copy()
    data._X = data._X[:, :1] # only use one feature

    subset, _ = data.split(test_size=0.98, random_state=random_state, stratify=True)
    train, test = subset.split(test_size=0.2, random_state=random_state, stratify=True)

    clf.fit(*train.X_y_lengths)

    k_idxs, k_distances, k_outputs = clf.query_neighbors(*test.X_lengths, sort=sort)

    # check that indices are between 0 and len(train)
    assert np.issubdtype(k_idxs.dtype, np.integer)
    assert k_idxs.shape == (len(test), clf.k)
    assert set(k_idxs.flatten()).issubset(set(np.arange(len(train))))

    # check that distances are sorted if sort=True
    np.issubdtype(k_distances.dtype, np.floating)
    assert k_distances.shape == (len(test), clf.k)
    if sort and k > 1:
        assert (k_distances[:, 1:] >= k_distances[:, :-1]).all()

    # check that labels are a subset of training outputs + check that outputs match indices
    assert np.issubdtype(k_outputs.dtype, np.floating)
    assert k_outputs.shape == (len(test), clf.k)
    assert set(k_outputs.flatten()).issubset(set(train.y))
    assert_equal(train.y[k_idxs], k_outputs)



from pkg_resources import resource_filename
from typing import Iterable, Tuple
from operator import itemgetter

import numpy as np
from pydantic import conint, validator
from sklearn.preprocessing import LabelEncoder

from sequentia.utils.data import SequentialDataset
from sequentia.utils.validation import _Validator
from sequentia.utils.decorators import _validate_params


class _GeneFamiliesValidator(_Validator):
    families: Iterable[conint(ge=0, le=6)] = list(range(7))

    @validator('families')
    def check_families(cls, value):
        value = list(value)
        if len(set(value)) < len(value):
            raise ValueError('Expected gene families to be unique')
        return value

@_validate_params(using=_GeneFamiliesValidator)
def load_gene_families(
    *,
    families: Iterable[int] = list(range(7))
) -> Tuple[SequentialDataset, LabelEncoder]:
    """Loads human DNA sequences grouped by gene family.

    The `Human DNA Sequences <https://www.kaggle.com/datasets/sooryaprakash12/human-dna-sequences>`_ dataset
    consists of 4380 DNA sequences belonging to 7 gene families.

    This dataset has imbalanced classes, and uses an :class:`sklearn:sklearn.preprocessing.LabelEncoder` to
    encode the original symbols (``A``, ``T``, ``C``, ``G``, ``N``) that form the DNA sequences, into integers.

    The gene families have the following class labels:

    - G protein coupled receptors: ``0``
    - Tyrosine kinase: ``1``
    - Tyrosine phosphatase: ``2``
    - Synthetase: ``3``
    - Synthase: ``4``
    - Ion channel: ``5``
    - Transcription: ``6``

    :param families: Subset of gene families to include in the dataset.

    :return:

        - A dataset object representing the loaded genetic data.
        - Label encoder used to encode the observation symbols into integers.
    """
    # Load the dataset from compressed numpy file
    data = np.load(resource_filename('sequentia', 'datasets/data/gene_families.npz'))

    # Fetch arrays from loaded file
    X, y, lengths = itemgetter('X', 'y', 'lengths')(data)

    # Encode the observation symbols into integers
    enc = LabelEncoder()
    X = np.expand_dims(enc.fit_transform(X.flatten()), axis=-1)

    # Select and create a Dataset only with sequences having the specified labels
    idx = np.argwhere(np.isin(y, families)).flatten()
    ranges = SequentialDataset._get_idxs(lengths)[idx]
    data = SequentialDataset(
        np.vstack([x for x in SequentialDataset._iter_X(X, ranges)]),
        y[idx],
        lengths[idx]
    )

    return data, enc



from pkg_resources import resource_filename
from typing import Iterable
from operator import itemgetter

import numpy as np
from pydantic import conint, validator

from sequentia.utils.data import SequentialDataset
from sequentia.utils.validation import _Validator
from sequentia.utils.decorators import _validate_params

class _DigitsValidator(_Validator):
    digits: Iterable[conint(ge=0, le=9)] = list(range(10))

    @validator('digits')
    def check_digits(cls, value):
        value = list(value)
        if len(set(value)) < len(value):
            raise ValueError('Expected digits to be unique')
        return value

@_validate_params(using=_DigitsValidator)
def load_digits(
    *,
    digits: Iterable[int] = list(range(10)),
) -> SequentialDataset:
    """Loads MFCC features of spoken digit audio samples from the Free Spoken Digit Dataset.

    The `Free Spoken Digit Dataset (FSDD) <https://github.com/Jakobovski/free-spoken-digit-dataset>`_
    consists of 3000 recordings of the spoken digits 0-9.

    This version consists of 13 MFCC features of 50 recordings for each digit by 6 individual speakers.

    :param digits: Subset of digits to include in the dataset.
    :return: A dataset object representing the loaded digits.
    """
    # Load the dataset from compressed numpy file
    data = np.load(resource_filename('sequentia', 'datasets/data/digits.npz'))

    # Fetch arrays from loaded file
    X, y, lengths = itemgetter('X', 'y', 'lengths')(data)

    # Select and create a Dataset only with sequences having the specified labels
    idx = np.argwhere(np.isin(y, digits)).flatten()
    ranges = SequentialDataset._get_idxs(lengths)[idx]
    return SequentialDataset(
        np.vstack([x for x in SequentialDataset._iter_X(X, ranges)]),
        y[idx],
        lengths[idx]
    )



import os
import pytest
from copy import deepcopy
from tempfile import TemporaryDirectory
from unittest.mock import Mock

import numpy as np

from sequentia.datasets import load_digits, load_gene_families
from sequentia.models.hmm import GaussianMixtureHMM, CategoricalHMM, HMMClassifier
from sequentia.utils.validation import _check_is_fitted

from .variants.test_gaussian_mixture import assert_fit as assert_gaussian_mixture_fit
from .variants.test_categorical import assert_fit as assert_categorical_fit
from ....support.assertions import assert_equal

n_classes = 7


@pytest.fixture(scope='module')
def random_state(request):
    return np.random.RandomState(1)


@pytest.fixture(scope='module')
def dataset(request):
    if request.param == 'digits':
        return load_digits(digits=range(n_classes))
    elif request.param == 'gene_families':
        data, _ = load_gene_families()
        return data


@pytest.fixture(scope='module')
def model(random_state, request):
    if request.param == 'gaussian_mixture':
        return GaussianMixtureHMM(topology='left-right', n_states=2, n_components=1, random_state=random_state)
    elif request.param == 'categorical':
        return CategoricalHMM(topology='left-right', n_states=2, random_state=random_state)


class MockData:
    def __init__(self, length):
        self.length = length

    def __len__(self):
        return self.length

    @property
    def lengths(self):
        return MockData(self.length)


def assert_fit(clf):
    assert hasattr(clf, 'prior_')
    assert hasattr(clf, 'classes_')
    assert _check_is_fitted(clf, return_=True)

    for hmm in clf.models.values():
        data = MockData(hmm.n_seqs_)
        if isinstance(hmm, GaussianMixtureHMM):
            assert_gaussian_mixture_fit(hmm, data)
        elif isinstance(hmm, CategoricalHMM):
            assert_categorical_fit(hmm, data)


@pytest.mark.parametrize(
    'model, dataset', [
        ('gaussian_mixture', 'digits'),
        ('categorical', 'gene_families')
    ],
    indirect=True
)
@pytest.mark.parametrize(
    'prior', [
        None,
        'frequency',
        {i: (i + 1) / (n_classes * (n_classes + 1) / 2) for i in range(n_classes)}
    ]
)
@pytest.mark.parametrize('prefit', [True, False])
def test_classifier_e2e(request, model, dataset, prior, prefit, random_state):
    clf = HMMClassifier(prior=prior)
    clf.add_models({i: deepcopy(model) for i in range(n_classes)})

    assert clf.prior == prior
    assert len(clf.models) == n_classes
    assert set(clf.models) == set(range(n_classes))
    assert all(isinstance(hmm, type(model)) for hmm in clf.models.values())

    subset, _ = dataset.split(test_size=0.6, random_state=random_state, stratify=True)
    train, test = subset.split(test_size=0.2, random_state=random_state, stratify=True)

    if prefit:
        for X, lengths, c in train.iter_by_class():
            clf.models[c].fit(X, lengths)
        assert_fit(clf.fit())
    else:
        assert_fit(clf.fit(*train.X_y_lengths))

    scores_pred = clf.predict_scores(*test.X_lengths)
    assert scores_pred.shape == (len(test), n_classes)

    proba_pred = clf.predict_proba(*test.X_lengths)
    assert proba_pred.shape == (len(test), n_classes)
    assert_equal(proba_pred.sum(axis=1), 1)
    assert ((proba_pred >= 0) & (proba_pred <= 1)).all()

    y_pred = clf.predict(*test.X_lengths)
    assert y_pred.shape == (len(test),)
    assert set(y_pred).issubset(set(range(n_classes)))

    acc = clf.score(*test.X_y_lengths)
    assert 0 <= acc <= 1

    # check serialization/deserialization
    with TemporaryDirectory() as temp_dir:
        model_path = f"{temp_dir}/{request.node.originalname}.model"
        # check that save works
        clf.save(model_path)
        assert os.path.isfile(model_path)
        # check that load works
        clf = HMMClassifier.load(model_path)
        # check that loaded model is fitted
        assert_fit(clf)
        y_pred_load = clf.predict(*test.X_lengths)
        # check that predictions are the same as before serialization
        assert_equal(y_pred, y_pred_load)


@pytest.mark.parametrize('classes', [[0, 1, 2], [2, 0, 1]])
def test_classifier_compute_log_posterior(classes):
    clf = HMMClassifier()
    clf.classes_ = np.array(classes)
    clf.prior_ = {i: np.exp(i) for i in clf.classes_}
    clf.models = {i: Mock(_score=Mock(side_effect=lambda x: 0)) for i in clf.classes_}
    assert_equal(clf._compute_log_posterior(None), clf.classes_)


def test_classifier_compute_scores_chunk():
    clf = HMMClassifier()
    clf.classes_ = np.arange(3)
    clf.prior_ = {i: np.exp(i) for i in clf.classes_}
    clf.models = {i: Mock(_score=Mock(side_effect=len)) for i in clf.classes_}
    X = np.expand_dims(np.arange(10), axis=-1)
    idxs = np.array([[0, 0], [1, 2], [3, 5], [6, 9]]) # lengths = 0, 1, 2, 3
    assert_equal(
        clf._compute_scores_chunk(idxs, X),
        np.tile(np.expand_dims(clf.classes_, axis=-1), len(idxs)).T
        + np.expand_dims(np.arange(len(idxs)), axis=-1)
    )



#!/usr/bin/env python
# -*- encoding: utf-8 -*-

import re
from setuptools import setup, find_packages
from pathlib import Path

with open('README.md', 'r', encoding='utf8') as fh:
    long_description = fh.read()

init = Path(__file__).parent / "lib" / "sequentia" / "__init__.py"
def load_meta(meta):
    with open(init, "r") as file:
        info = re.search(rf'^__{meta}__\s*=\s*[\'"]([^\'"]*)[\'"]', file.read(), re.MULTILINE).group(1)
        if not info:
            raise RuntimeError(f"Could not load {repr(meta)} metadata")
        return info

setup(
    name = load_meta("name"),
    version = load_meta("version"),
    author = load_meta("author"),
    author_email = load_meta("email"),
    description = 'HMM and DTW-based sequence machine learning algorithms in Python following an sklearn-like interface.',
    long_description = long_description,
    long_description_content_type = 'text/markdown',
    url = 'https://github.com/eonu/sequentia',
    project_urls = {
        'Documentation': 'https://sequentia.readthedocs.io/en/latest',
        'Bug Tracker': 'https://github.com/eonu/sequentia/issues',
        'Source Code': 'https://github.com/eonu/sequentia',
    },
    license = 'MIT',
    package_dir = {'': 'lib'},
    packages = find_packages(where='lib'),
    package_data={
        'sequentia': [
            'datasets/data/digits.npz',
            'datasets/data/gene_familites.npz',
        ]
    },
    classifiers = [
        'Development Status :: 5 - Production/Stable',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'License :: OSI Approved :: MIT License',
        'Operating System :: Unix',
        'Operating System :: MacOS',
        'Intended Audience :: Science/Research',
        'Topic :: Software Development',
        'Topic :: Scientific/Engineering',
        'Natural Language :: English',
    ],
    python_requires = '>=3.8',
    setup_requires = [
        'Cython>=0.28.5',
        'numpy>=1.18,<1.24',
        'scipy>=1.3',
    ],
    install_requires = [
        'numba>=0.56',
        'numpy>=1.18,<1.24',
        'hmmlearn>=0.2.8',
        'dtaidistance>=2.3.10', # [numpy]
        'scikit-learn>=1.0',
        'joblib>=0.14',
        'pydantic<1.9',
    ],
    extras_require = {
        'dev': [
            'sphinx',
            'numpydoc',
            'sphinx_rtd_theme',
            'sphinx-autodoc-typehints',
            'sphinx-autobuild',
            'm2r2',
            'mistune==0.8.4',
            'Jinja2',
            'pytest',
        ]
    }
)



from __future__ import annotations

import re
import warnings
from copy import deepcopy
from types import SimpleNamespace
from typing import Optional, Union, Dict, Any, Literal
from pydantic import NonNegativeInt, PositiveInt, validator

import numpy as np
from sklearn.base import BaseEstimator
from sklearn.utils import check_random_state

from sequentia.utils.decorators import _requires_fit
from sequentia.utils.validation import Array, _Validator
from sequentia.models.hmm.topologies import _ErgodicTopology

_defaults = SimpleNamespace(
    n_states=5,
    topology="left-right",
    random_state=None,
    hmmlearn_kwargs=dict(
        init_params="st",
        params="st",
    ),
)

class _HMM(BaseEstimator):
    _base_sequence_validator = None
    _single_sequence_validator = None
    _sequence_classifier_validator = None
    _defaults = _defaults
    _unsettable_hmmlearn_kwargs = ["random_state", "init_params", "params"]

    def __init__(
        self,
        n_states: PositiveInt,
        topology: Optional[str],
        random_state: Optional[Union[NonNegativeInt, np.random.RandomState]],
        hmmlearn_kwargs: Dict[str, Any]
    ) -> _HMM:
        if type(self) == _HMM:
            raise NotImplementedError(
                f'Abstract class {type(self).__name__} cannot be instantiated - '
                'use the subclassing HMMs defined in the sequentia.models.hmm module'
            )

        #: Number of states in the Markov chain.
        self.n_states = n_states
        #: Transition topology of the Markov chain — see :ref:`topologies`.
        self.topology = topology
        #: Seed or :class:`numpy:numpy.random.RandomState` object for reproducible pseudo-randomness.
        self.random_state = random_state
        #: Additional key-word arguments provided to the `hmmlearn <https://hmmlearn.readthedocs.io/en/latest/>`__ HMM constructor.
        self.hmmlearn_kwargs = deepcopy(hmmlearn_kwargs)
        #: Underlying HMM object from `hmmlearn <https://hmmlearn.readthedocs.io/en/latest/>`__ — only set after :func:`fit`.
        self.model = None

        self._skip_init_params = set()
        self._skip_params = set()

    def fit(
        self,
        X: Array,
        lengths: Optional[Array[int]] = None
    ) -> _HMM:
        raise NotImplementedError

    @_requires_fit
    def score(
        self,
        x: Array,
    ) -> float:
        data = self._single_sequence_validator(sequence=x)
        return self._score(data.sequence)

    @_requires_fit
    def n_params(self) -> NonNegativeInt:
        """Retrieves the number of trainable parameters.

        :note: This method requires a trained model — see :func:`fit`.

        :return: Number of trainable parameters.
        """
        n_params = 0
        if 's' not in self._skip_params:
            n_params += self.model.startprob_.size
        if 't' not in self._skip_params:
            n_params += self.model.transmat_.size
        return n_params

    @_requires_fit
    def bic(
        self,
        X: Array,
        lengths: Optional[Array[int]] = None
    ) -> float:
        data = self._base_sequence_validator(X=X, lengths=lengths)
        max_log_likelihood = self.model.score(data.X, data.lengths)
        n_params = self.n_params()
        n_seqs = len(lengths)
        return n_params * np.log(n_seqs) - 2 * np.log(max_log_likelihood)

    @_requires_fit
    def aic(
        self,
        X: Array,
        lengths: Optional[Array[int]] = None
    ) -> float:
        data = self._base_sequence_validator(X=X, lengths=lengths)
        max_log_likelihood = self.model.score(data.X, data.lengths)
        n_params = self.n_params()
        return 2 * n_params - 2 * np.log(max_log_likelihood)

    def set_start_probs(
        self,
        values: Union[Array, Literal["uniform", "random"]] = 'random'
    ):
        """Sets the initial state probabilities.

        If this method is **not** called, initial state probabilities are initialized depending on the value of ``topology`` provided to :func:`__init__`.

        - If ``topology`` was set to ``'ergodic'``, ``'left-right'`` or ``'linear'``, then random probabilities will be assigned according to the topology by calling :func:`set_start_probs` with ``value='random'``.
        - If ``topology`` was set to ``None``, then initial state probabilities will be initialized by `hmmlearn <https://hmmlearn.readthedocs.io/en/latest/>`__.

        :param values: Probabilities or probability type to assign as initial state probabilities.

            - If an ``Array``, should be a vector of starting probabilities for each state.
            - If ``'uniform'``, there is an equal probability of starting in any state.
            - If ``'random'``, the vector of initial state probabilities is sampled
              from a Dirichlet distribution with unit concentration parameters.

        :note: If used, this method should normally be called before :func:`fit`.
        """
        error = ValueError("Invalid start probabilities - expected: 'uniform', 'random' or an array of probabilities")
        if isinstance(values, str):
            if values in ('uniform', 'random'):
                self._startprob = values
                self._skip_init_params |= set('s')
            else:
                raise error
        else:
            try:
                self._startprob = np.array(values)
                self._skip_init_params |= set('s')
            except Exception as e:
                raise error from e

    def set_transitions(
        self,
        values: Union[Array, Literal["uniform", "random"]] = 'random'
    ):
        """Sets the transition probability matrix.

        If this method is **not** called, transition probabilities are initialized depending on the value of ``topology`` provided to :func:`__init__`:

        - If ``topology`` was set to ``'ergodic'``, ``'left-right'`` or ``'linear'``, then random probabilities will be assigned according to the topology by calling :func:`set_transitions` with ``value='random'``.
        - If ``topology`` was set to ``None``, then initial state probabilities will be initialized by `hmmlearn <https://hmmlearn.readthedocs.io/en/latest/>`__.

        :param values: Probabilities or probability type to assign as state transition probabilities.

            - If an ``Array``, should be a matrix of probabilities where each row must some to one
              and represents the probabilities of transitioning out of a state.
            - If ``'uniform'``, for each state there is an equal probability of transitioning to any state permitted by the topology.
            - If ``'random'``, the vector of transition probabilities for each row is sampled from a
              Dirichlet distribution with unit concentration parameters, according to the shape of the topology.

        :note: If used, this method should normally be called before :func:`fit`.
        """
        error = ValueError("Invalid transition matrix - expected: 'uniform', 'random' or an array of probabilities")
        if isinstance(values, str):
            if values in ('uniform', 'random'):
                self._transmat = values
                self._skip_init_params |= set('t')
            else:
                raise error
        else:
            try:
                self._transmat = np.array(values)
                self._skip_init_params |= set('t')
            except Exception as e:
                raise error from e

    def freeze(
        self,
        params: str,
    ):
        self._skip_params |= set(self._modify_params(params))

    def unfreeze(
        self,
        params: str,
    ):
        self._skip_params -= set(self._modify_params(params))

    def _modify_params(self, params):
        defaults = deepcopy(self._defaults.hmmlearn_kwargs["params"])
        error_msg = f"Expected a string consisting of any combination of {defaults}"
        if isinstance(params, str):
            if bool(re.compile(fr'[^{defaults}]').search(params)):
                raise ValueError(error_msg)
        else:
            raise TypeError(error_msg)
        return params

    def _check_init_params(self):
        topology = self.topology_ or _ErgodicTopology(self.n_states, check_random_state(self.random_state))

        if 's' in self._skip_init_params:
            if isinstance(self._startprob, str):
                if self._startprob == 'uniform':
                    self._startprob = topology.uniform_start_probs()
                elif self._startprob == 'random':
                    self._startprob = topology.random_start_probs()
            elif isinstance(self._startprob, np.ndarray):
                self._startprob = topology.check_start_probs(self._startprob)
        else:
            if self.topology_ is not None:
                self.set_start_probs(topology.random_start_probs())

        if 't' in self._skip_init_params:
            if isinstance(self._transmat, str):
                if self._transmat == 'uniform':
                    self._transmat = topology.uniform_transitions()
                elif self._transmat == 'random':
                    self._transmat = topology.random_transitions()
            elif isinstance(self._transmat, np.ndarray):
                self._transmat = topology.check_transitions(self._transmat)
        else:
            if self.topology_ is not None:
                self.set_transitions(topology.random_transitions())

    def _score(self, x: Array) -> float:
        return self.model.score(x)

class _HMMValidator(_Validator):
    n_states: PositiveInt = _defaults.n_states
    topology: Optional[Literal["ergodic", "left-right", "linear"]] = _defaults.topology
    random_state: Optional[Union[NonNegativeInt, np.random.RandomState]] = _defaults.random_state
    hmmlearn_kwargs: Dict[str, Any] = deepcopy(_defaults.hmmlearn_kwargs)

    _class = _HMM

    @validator('random_state')
    def check_random_state(cls, value):
        return check_random_state(value)

    @validator('hmmlearn_kwargs')
    def check_hmmlearn_kwargs(cls, value):
        params = deepcopy(value)

        defaults = deepcopy(cls._class._defaults.hmmlearn_kwargs["params"])
        setter_methods = [f"{func}()" for func in dir(cls._class) if func.startswith("set") and func != "set_params"]

        for param in value.keys():
            if param in cls._class._unsettable_hmmlearn_kwargs:
                if param == 'init_params':
                    if set(params[param]) != set(defaults):
                        params[param] = defaults
                        warnings.warn(
                            f"The `init_params` hmmlearn argument cannot be overridden manually - defaulting to all parameters '{defaults}'. "
                            f'Use the following methods to initialize model parameters: {", ".join(setter_methods)}.'
                        )
                elif param == 'params':
                    if set(params[param]) != set(defaults):
                        params[param] = defaults
                        warnings.warn(
                            f"The `params` hmmlearn argument cannot be overridden manually - defaulting to all parameters '{defaults}'. "
                            'Use the freeze() and unfreeze() methods to specify the learnable model parameters.'
                        )
                else:
                    del params[param]
                    warnings.warn(
                        f'The `{param}` hmmlearn argument cannot be overriden manually - use the {cls._class.__name__} constructor to specify this argument.'
                    )

        if 'init_params' not in params:
            params['init_params'] = defaults
            warnings.warn(
                f"No initializable parameters set in hmmlearn `init_params` argument - defaulting to '{defaults}'. "
                f'If you intend to manually initialize all parameters, use the following methods: {", ".join(setter_methods)}.'
            )

        if 'params' not in params:
            params['params'] = defaults
            warnings.warn(
                f"No learnable parameters set in hmmlearn `params` argument - defaulting to '{defaults}'. "
                'If you intend to make no parameters learnable, use the freeze() method.'
            )

        return params


import inspect, functools

from sequentia.utils.validation import _check_is_fitted

def _validate_params(using):
    def decorator(function):
        @functools.wraps(function)
        def wrapper(self=None, *args, **kwargs):
            spec = inspect.getfullargspec(function)
            if spec.varkw == 'kwargs' or len(spec.kwonlyargs) > 0:
                using.parse_obj(kwargs)
            if self is None:
                return function(*args, **kwargs)
            return function(self, *args, **kwargs)
        return wrapper
    return decorator

def _requires_fit(function):
    @functools.wraps(function)
    def wrapper(self, *args, **kwargs):
        _check_is_fitted(self)
        return function(self, *args, **kwargs)
    return wrapper

def _override_params(params, temporary=True):
    def decorator(function):
        @functools.wraps(function)
        def wrapper(self, *args, **kwargs):
            original_params = {}

            for param in params:
                if not hasattr(self, param):
                    raise AttributeError(f"'{type(self).__name__}' object has no attribute '{param}'")

            for param in params:
                if param in kwargs:
                    original_params[param] = getattr(self, param)
                    setattr(self, param, kwargs[param])

            try:
                return function(self, *args, **kwargs)
            finally:
                if temporary:
                    for param, value in original_params.items():
                        setattr(self, param, value)

        return wrapper
    return decorator

def _check_plotting_dependencies(function):
    @functools.wraps(function)
    def wrapper(self, *args, **kwargs):
        try:
            import matplotlib
        except ImportError as e:
            raise ImportError(f'The {function.__name__} function requires a working installation of matplotlib') from e
        return function(self, *args, **kwargs)
    return wrapper



from __future__ import annotations

from copy import deepcopy
from types import SimpleNamespace
from typing import Optional, Union, Dict, Any, Literal
from pydantic import NonNegativeInt, PositiveInt

import numpy as np
import hmmlearn.hmm
from sklearn.utils import check_random_state

from sequentia.models.hmm.topologies import _topologies
from sequentia.models.hmm.variants.base import _HMM, _HMMValidator
from sequentia.utils.decorators import _validate_params, _requires_fit
from sequentia.utils.validation import (
    Array,
    _BaseMultivariateFloatSequenceValidator,
    _SingleMultivariateFloatSequenceValidator,
    _MultivariateFloatSequenceClassifierValidator,
)

__all__ = ['GaussianMixtureHMM']

_defaults = SimpleNamespace(
    **{
        **_HMM._defaults.__dict__,
        "n_components": 3,
        "covariance_type": "spherical",
        "hmmlearn_kwargs": dict(
            init_params="stmcw",
            params="stmcw",
        )
    }
)


class GaussianMixtureHMM(_HMM):
    """A hidden Markov model with multivariate Gaussian mixture emissions.

    Examples
    --------
    Using a :class:`.GaussianMixtureHMM` to learn how to recognize spoken samples of the digit 3.

    See :func:`.load_digits` for more information on the sample dataset used in this example.

    ::

        import numpy as np
        from sequentia.datasets import load_digits
        from sequentia.models.hmm import GaussianMixtureHMM

        # Seed for reproducible pseudo-randomness
        random_state = np.random.RandomState(1)

        # Fetch MFCCs of spoken samples for the digit 3
        data = load_digits(digits=[3])
        train_data, test_data = data.split(test_size=0.2, random_state=random_state)

        # Create and train a GaussianMixtureHMM to recognize the digit 3
        model = GaussianMixtureHMM(random_state=random_state)
        X_train, lengths_train = train_data.X_lengths
        model.fit(X_train, lengths_train)

        # Calculate the log-likelihood of the first test sample being generated by this model
        x, y = test_data[0]
        model.score(x)
    """

    _base_sequence_validator = _BaseMultivariateFloatSequenceValidator
    _single_sequence_validator = _SingleMultivariateFloatSequenceValidator
    _sequence_classifier_validator = _MultivariateFloatSequenceClassifierValidator
    _defaults = _defaults
    _unsettable_hmmlearn_kwargs = _HMM._unsettable_hmmlearn_kwargs + ["n_components", "n_mix", "covariance_type"]

    def __init__(
        self,
        *,
        n_states: PositiveInt = _defaults.n_states,
        n_components: PositiveInt = _defaults.n_components,
        covariance_type: Literal["spherical", "diag", "full", "tied"] = _defaults.covariance_type,
        topology: Optional[Literal["ergodic", "left-right", "linear"]] = _defaults.topology,
        random_state: Optional[Union[NonNegativeInt, np.random.RandomState]] = _defaults.random_state,
        hmmlearn_kwargs: Dict[str, Any] = deepcopy(_defaults.hmmlearn_kwargs)
    ) -> GaussianMixtureHMM:
        """Initializes the :class:`.GaussianMixtureHMM`.

        :param n_states: Number of states in the Markov chain.
        :param n_components: Number of Gaussian components in the mixture emission distribution for each state.
        :param covariance_type: Type of covariance matrix in the mixture emission distribution for each state - see :ref:`covariance_types`.
        :param topology: Transition topology of the Markov chain — see :ref:`topologies`.

            - If ``None``, behaves the same as ``'ergodic'`` but with `hmmlearn <https://hmmlearn.readthedocs.io/en/latest/>`__ initialization.

        :param random_state: Seed or :class:`numpy:numpy.random.RandomState` object for reproducible pseudo-randomness.
        :param hmmlearn_kwargs: Additional key-word arguments provided to the `hmmlearn <https://hmmlearn.readthedocs.io/en/latest/>`__ HMM constructor.
        """
        super().__init__(n_states, topology, random_state, hmmlearn_kwargs)
        #: Number of Gaussian components in the emission model mixture distribution for each state.
        self.n_components = n_components
        #: Type of covariance matrix in the emission model mixture distribution for each state.
        self.covariance_type = covariance_type

    def fit(
        self,
        X: Array[float],
        lengths: Optional[Array[int]] = None
    ) -> GaussianMixtureHMM:
        """Fits the HMM to the sequences in ``X``, using the Baum—Welch algorithm.

        :param X: Univariate or multivariate observation sequence(s).

            - Should be a single 1D or 2D array.
            - Should have length as the 1st dimension and features as the 2nd dimension.
            - Should be a concatenated sequence if multiple sequences are provided,
              with respective sequence lengths being provided in the ``lengths`` argument for decoding the original sequences.

        :param lengths: Lengths of the observation sequence(s) provided in ``X``.

            - If ``None``, then ``X`` is assumed to be a single observation sequence.
            - ``len(X)`` should be equal to ``sum(lengths)``.

        :return: The fitted HMM.
        """
        data = self._base_sequence_validator(X=X, lengths=lengths)
        self.random_state_ = check_random_state(self.random_state)
        if self.topology is None:
            self.topology_ = None
        else:
            self.topology_ = _topologies[self.topology](self.n_states, self.random_state_)
        self._check_init_params()

        kwargs = deepcopy(self.hmmlearn_kwargs)
        kwargs['init_params'] = ''.join(set(kwargs['init_params']) - self._skip_init_params)
        kwargs['params'] = ''.join(set(kwargs['params']) - self._skip_params)

        self.model = hmmlearn.hmm.GMMHMM(
            n_components=self.n_states,
            n_mix=self.n_components,
            covariance_type=self.covariance_type,
            random_state=self.random_state_,
            **kwargs
        )

        for attr in ('startprob', 'transmat', 'means', 'covars', 'weights'):
            if hasattr(self, f'_{attr}'):
                setattr(self.model, f'{attr}_', getattr(self, f'_{attr}'))

        self.model.fit(data.X, lengths=data.lengths)
        self.n_seqs_ = len(data.lengths)

        return self

    @_requires_fit
    def score(
        self,
        x: Array[float],
    ) -> float:
        """Calculates the log-likelihood of the HMM generating a single observation sequence.

        :param x: Univariate or multivariate observation sequence.

            - Should be a single 1D or 2D array.
            - Should have length as the 1st dimension and features as the 2nd dimension.

        :note: This method requires a trained model — see :func:`fit`.

        :return: The log-likelihood.
        """
        return super().score(x)

    @_requires_fit
    def n_params(self) -> NonNegativeInt:
        n_params = super().n_params()
        if 'm' not in self._skip_params:
            n_params += self.model.means_.size
        if 'c' not in self._skip_params:
            n_params += self.model.covars_.size
        if 'w' not in self._skip_params:
            n_params += self.model.weights_.size
        return n_params

    @_requires_fit
    def bic(
        self,
        X: Array[float],
        lengths: Optional[Array[int]] = None
    ) -> float:
        """The Bayesian information criterion of the model, evaluated with the maximum likelihood of ``X``.

        :param X: Univariate or multivariate observation sequence(s).

            - Should be a single 1D or 2D array.
            - Should have length as the 1st dimension and features as the 2nd dimension.
            - Should be a concatenated sequence if multiple sequences are provided,
              with respective sequence lengths being provided in the ``lengths`` argument for decoding the original sequences.

        :param lengths: Lengths of the observation sequence(s) provided in ``X``.

            - If ``None``, then ``X`` is assumed to be a single observation sequence.
            - ``len(X)`` should be equal to ``sum(lengths)``.

        :note: This method requires a trained model — see :func:`fit`.

        :return: The Bayesian information criterion.
        """
        return super().bic(X, lengths)

    @_requires_fit
    def aic(
        self,
        X: Array[float],
        lengths: Optional[Array[int]] = None
    ) -> float:
        """The Akaike information criterion of the model, evaluated with the maximum likelihood of ``X``.

        :param X: Univariate or multivariate observation sequence(s).

            - Should be a single 1D or 2D array.
            - Should have length as the 1st dimension and features as the 2nd dimension.
            - Should be a concatenated sequence if multiple sequences are provided,
              with respective sequence lengths being provided in the ``lengths`` argument for decoding the original sequences.

        :param lengths: Lengths of the observation sequence(s) provided in ``X``.

            - If ``None``, then ``X`` is assumed to be a single observation sequence.
            - ``len(X)`` should be equal to ``sum(lengths)``.

        :note: This method requires a trained model — see :func:`fit`.

        :return: The Akaike information criterion.
        """
        return super().aic(X, lengths)

    def set_state_means(self, values: Array[float]):
        """Sets the mean vectors of the state emission distributions.

        If this method is **not** called, mean vectors will be initialized by `hmmlearn <https://hmmlearn.readthedocs.io/en/latest/>`__.

        :param values: Array of emission distribution mean values.

        :note: If used, this method should normally be called before :func:`fit`.
        """
        self._means = Array[float].validate_type(values)
        self._skip_init_params |= set('m')

    def set_state_covariances(self, values: Array[float]):
        """Sets the covariance matrices of the state emission distributions.

        If this method is **not** called, covariance matrices will be initialized by `hmmlearn <https://hmmlearn.readthedocs.io/en/latest/>`__.

        :param values: Array of emission distribution covariance values.

        :note: If used, this method should normally be called before :func:`fit`.
        """
        self._covars = Array[float].validate_type(values)
        self._skip_init_params |= set('c')

    def set_state_weights(self, values: Array[float]):
        """Sets the component mixture weights of the state emission distributions.

        If this method is **not** called, component mixture weights will be initialized by `hmmlearn <https://hmmlearn.readthedocs.io/en/latest/>`__.

        :param values: Array of emission distribution component mixture weights.

        :note: If used, this method should normally be called before :func:`fit`.
        """
        self._weights = Array[float].validate_type(values)
        self._skip_init_params |= set('w')

    def freeze(
        self,
        params: str = deepcopy(_defaults.hmmlearn_kwargs["params"]),
    ):
        """Freezes the trainable parameters of the HMM, preventing them from being updated during the Baum—Welch algorithm.

        :param params: A string specifying which parameters to freeze. Can contain a combination of:

            - ``'s'`` for initial state probabilities,
            - ``'t'`` for transition probabilities,
            - ``'m'`` for emission distribution means,
            - ``'c'`` for emission distribution covariances,
            - ``'w'`` for emission distribution mixture weights.

        :note: If used, this method should normally be called before :func:`fit`.

        See Also
        --------
        unfreeze:
            Unfreezes the trainable parameters of the HMM, allowing them to be updated during the Baum—Welch algorithm.
        """
        super().freeze(params)

    def unfreeze(
        self,
        params: str = deepcopy(_defaults.hmmlearn_kwargs["params"]),
    ):
        """Unfreezes the trainable parameters of the HMM, allowing them to be updated during the Baum—Welch algorithm.

        :param params: A string specifying which parameters to unfreeze. Can contain a combination of:

            - ``'s'`` for initial state probabilities,
            - ``'t'`` for transition probabilities,
            - ``'m'`` for emission distribution means,
            - ``'c'`` for emission distribution covariances,
            - ``'w'`` for emission distribution mixture weights.

        See Also
        --------
        freeze:
            Freezes the trainable parameters of the HMM, preventing them from be updated during the Baum—Welch algorithm.
        """
        super().unfreeze(params)

class _GaussianMixtureHMMValidator(_HMMValidator):
    n_components: PositiveInt = _defaults.n_components
    covariance_type: Literal["spherical", "diag", "full", "tied"] = _defaults.covariance_type
    hmmlearn_kwargs: Dict[str, Any] = deepcopy(_defaults.hmmlearn_kwargs)

    _class = GaussianMixtureHMM

GaussianMixtureHMM.__init__ = _validate_params(using=_GaussianMixtureHMMValidator)(GaussianMixtureHMM.__init__)



import os
import pytest
from tempfile import TemporaryDirectory

import numpy as np

from sequentia.utils import SequentialDataset

from ...support.assertions import assert_equal, assert_all_equal


@pytest.mark.parametrize('y_type', [int, float, None])
@pytest.mark.parametrize('use_lengths', [True, False])
def test_data(request, y_type, use_lengths):
    X = np.arange(10)

    if y_type == int:
        y = [10, 15, 10]
    elif y_type == float:
        y = [10.1, 15.2, 20.3]
    elif y_type is None:
        y = None

    if use_lengths:
        lengths = [2, 3, 5]
    else:
        lengths = None
        if y_type:
            y = y[0]

    data = SequentialDataset(X, y, lengths)

    # X
    assert_equal(data.X, np.atleast_2d(X).T)

    # y, classes
    if y_type == int:
        assert np.issubdtype(data.y.dtype, np.integer)
        assert_equal(data.y, np.atleast_1d(y))
        assert_equal(data.classes, [10, 15] if lengths else [10])
    elif y_type == float:
        assert np.issubdtype(data.y.dtype, np.floating)
        assert_equal(data.y, np.atleast_1d(y))
        assert data.classes is None
    elif y_type is None:
        for prop in ('y', 'X_y', 'X_y_lengths'):
            with pytest.raises(AttributeError):
                getattr(data, prop)
        assert data.classes is None

    # idxs
    if lengths:
        assert_equal(
            data.idxs, [
                [0,  2],
                [2,  5],
                [5, 10]
            ]
        )
    else:
        assert_equal(data.idxs, [[0, 10]])

    # _iter_X
    assert_equal(data.X, np.vstack([x for x in data._iter_X(data.X, data.idxs)]))

    # __getitem__
    if y_type:
        if lengths:
            # [0]
            x, y_ = data[0]
            assert_equal(x, np.atleast_2d([0, 1]).T)
            assert y_ == y[0]
            # [:1]
            xs, ys = data[:1]
            assert_all_equal(xs, [np.atleast_2d([0, 1]).T])
            assert_equal(ys, y[:1])
            # [1:3]
            xs, ys = data[1:3]
            assert_all_equal(xs, [
                np.atleast_2d([2, 3, 4]).T,
                np.atleast_2d([5, 6, 7, 8, 9]).T
            ])
            assert_equal(ys, y[1:3])
            # [-1]
            x, y_ = data[-1]
            assert_equal(x, np.atleast_2d([5, 6, 7, 8, 9]).T)
            assert y_ == y[-1]
            # [-2:]
            xs, ys = data[-2:]
            assert_all_equal(xs, [
                np.atleast_2d([2, 3, 4]).T,
                np.atleast_2d([5, 6, 7, 8, 9]).T
            ])
            assert_equal(ys, y[-2:])
        else:
            # [0]
            x, y_ = data[0]
            assert_equal(x, np.atleast_2d(X).T)
            assert y_ == y
            # [:1]
            xs, ys = data[:1]
            assert_all_equal(xs, [np.atleast_2d(X).T])
            assert_equal(ys, [y])
    else:
        if lengths:
            # [0]
            x = data[0]
            assert_equal(x, np.atleast_2d([0, 1]).T)
            # [:1]
            xs = data[:1]
            assert_all_equal(xs, [np.atleast_2d([0, 1]).T])
            # [1:3]
            xs = data[1:3]
            assert_all_equal(xs, [
                np.atleast_2d([2, 3, 4]).T,
                np.atleast_2d([5, 6, 7, 8, 9]).T
            ])
            # [-1]
            x = data[-1]
            assert_equal(x, np.atleast_2d([5, 6, 7, 8, 9]).T)
            # [-2:]
            xs = data[-2:]
            assert_all_equal(xs, [
                np.atleast_2d([2, 3, 4]).T,
                np.atleast_2d([5, 6, 7, 8, 9]).T
            ])
        else:
            # [0]
            x = data[0]
            assert_equal(x, np.atleast_2d(X).T)
            # [:1]
            xs = data[:1]
            assert_all_equal(xs, [np.atleast_2d(X).T])

    # split
    if y and lengths:
        train, test = data.split(test_size=1/3, shuffle=False)
        assert len(train) == 2
        assert len(test) == 1
        assert_equal(train.lengths, data.lengths[:len(train)])
        assert_equal(test.lengths, data.lengths[-len(test):])
        assert_equal(train.X, data.X[:train.lengths.sum()])
        assert_equal(test.X, data.X[-test.lengths.sum():])
        assert_equal(train.y, data.y[:len(train)])
        assert_equal(test.y, data.y[-len(test):])

    # iter_by_class
    if y_type == int and lengths:
        for X_, lengths_, c in data.iter_by_class():
            if c == 10:
                assert_equal(lengths_, [2, 5])
                assert_equal(X_, np.vstack([data.X[:2], data.X[-5:]]))
            elif c == 15:
                assert_equal(lengths_, [3])
                assert_equal(X_, data.X[2:5])

    # check serialization/deserialization
    with TemporaryDirectory() as temp_dir:
        data_path = f"{temp_dir}/{request.node.originalname}.npz"
        # check that save works
        data.save(data_path)
        assert os.path.isfile(data_path)
        # check that load works
        data_load = SequentialDataset.load(data_path)
        # check that stored values are the same
        assert_equal(data._X, data_load._X)
        assert_equal(data._lengths, data_load._lengths)
        if y:
            assert_equal(data._y, data_load._y)
        else:
            assert data._y is None and data_load._y is None
        if data._classes is not None:
            assert_equal(data._classes, data_load._classes)
        else:
            assert data._classes is None and data_load._classes is None

    # copy - check that stored values are the same
    data_copy = data.copy()
    assert_equal(data._X, data_copy._X)
    assert_equal(data._lengths, data_copy._lengths)
    if y:
        assert_equal(data._y, data_copy._y)
    else:
        assert data._y is None and data_copy._y is None
    if data._classes is not None:
        assert_equal(data._classes, data_copy._classes)
    else:
        assert data._classes is None and data_copy._classes is None



import pytest

import numpy as np

from sklearn.preprocessing import minmax_scale

from sequentia.datasets import load_digits
from sequentia.preprocessing import transforms

from sequentia.utils import SequentialDataset

from ...support.assertions import assert_equal


@pytest.fixture(scope='module')
def random_state(request):
    return np.random.RandomState(1)


@pytest.fixture(scope='module')
def data(random_state):
    data_= load_digits(digits=[0])
    _, subset = data_.split(test_size=0.2, random_state=random_state, stratify=True)
    return subset


def check_filter(x, xt, func, k):
    """Only works for odd k"""
    assert len(x) == len(xt)
    assert_equal(xt[k // 2], func(x[:k], axis=0))


def test_function_transformer(data):
    # create the transform
    transform = transforms.IndependentFunctionTransformer(minmax_scale)
    # check that fit works - should do nothing
    transform.fit(*data.X_lengths)
    # check that fit_transform works - shouldn't do anything on fit, but should transform
    X_fit_transform = transform.fit_transform(*data.X_lengths)
    # check that transform works
    X_transform = transform.transform(*data.X_lengths)
    # check that fit_transform and transform produce the same transformed data
    assert_equal(X_fit_transform, X_transform)
    # check that features of each sequence are independently scaled to [0, 1]
    for xt in SequentialDataset._iter_X(X_transform, data.idxs):
        assert_equal(xt.min(axis=0), np.zeros(xt.shape[1]))
        assert_equal(xt.max(axis=0), np.ones(xt.shape[1]))


@pytest.mark.parametrize("avg", ["mean", "median"])
@pytest.mark.parametrize("k", [3, 5])
def test_filters(data, random_state, avg, k):
    filter_ = getattr(transforms, f"{avg}_filter")
    check_filter_ = lambda x, xt: check_filter(x, xt, getattr(np, avg), k)

    # check that filters are correctly applied for a single sequence
    n_features = 2
    x = random_state.rand(10 * n_features).reshape(-1, n_features)
    xt = filter_(x, k)
    check_filter_(x, xt)

    # create a transform using the filter, passing k
    transform = transforms.IndependentFunctionTransformer(filter_, kw_args={"k": k})
    Xt = transform.transform(data.X, data.lengths)

    # check that filters are correctly applied for multiple sequences
    idxs = SequentialDataset._get_idxs(data.lengths)
    for x, xt in zip(*map(lambda X: SequentialDataset._iter_X(X, idxs), (data.X, Xt))):
        check_filter_(x, xt)
